{-# OPTIONS_GHC -fallow-undecidable-instances #-}
module Engine.RelOld where

import AST.Vector
import AST.Prolog
import AST.Relational
import Engine.Rewritable
import Engine.Unification (unify, apply)

import Compiler.Module

import Data.List
import Control.Monad.Trans
import Control.Monad.State
import Debug.Trace

type RModule = Module Predicate Relational

instance (MonadIO m, MonadState RModule m) => Rewritable m Relational where
    isFinal = relIsFinal
    rewriteStep = relRewrite

-- This module is at early stage of implementation.

debug s = liftIO $ putStr s

-- See AST.Relational about invariants that hold here.

relIsFinal :: Monad m => Relational -> m Bool
{-- An union is final if any member of it it's too.
relIsFinal (CUnion l) =
    do lbool <- mapM (relIsFinal) l
       return $ foldl (||) False lbool
-- For an intersection all elems must be final.
-- egallego@babel: Ummm, maybe an intersection is never final.
relIsFinal (CIntersection l) =
    do lbool <- mapM (relIsFinal) l
       return $ foldl (&&) True lbool
-- A concat: Have to think.
relIsFinal (CConcat l)   =
    do lbool <- mapM (relIsFinal) l
       return $ foldl (&&) True lbool
-- Every function is final except a Predicate.
relIsFinal (FPred s) = return False -}
relIsFinal (CUnion ((FK _):_)) = return True
relIsFinal (FK _   ) = return True
relIsFinal (F0) = return True
relIsFinal _         = return False

-- We filter any list of a single elem
--filterl [x] = x
--filterl l   = l

-- Here we model the rewrite rules as pattern matching in Haskell.

relRewrite                   :: (MonadIO m, MonadState RModule m) => Relational -> m Relational
relRewrite (CUnion l)        = rewriteUnion l
relRewrite (CIntersection l) = rewriteInter l
relRewrite (CConcat l)       = rewriteConcat l
relRewrite (FPred p )        =
    do m <- get
       case lookupCode m p of
            (newexpr:_) -> return newexpr
            []          -> return F0
relRewrite x                 = do 
--    debug $ "*** rewriting " ++ (show x) ++ "\n"
    return x

-- First we try to rewrite the known cases, then we continue.

rewriteConcat []  = return $ CConcat []
rewriteConcat [x] = return x
rewriteConcat [FI n, FK v, FI n'] =
    let v' = freshAndReduce n v in
         return $ FK v'

rewriteConcat [FConv (FW po), FK v, FW o] =
    let newk = permute v (invert o) in
        do --debug $ "\nPermute \n ko: " ++ (show v) ++ "\n\nkn: " ++ (show newk) ++ "\n\nConv : " ++ (show o) ++"\n\n"
          return $ FK newk

rewriteConcat [FW po, FK v, FConv(FW o)] =
    let newk = permute v po in
           return $ FK newk

rewriteConcat [FW po, CUnion l, FConv(FW o)] = do
    let newl = map (\e -> CConcat [FW po, e, FConv(FW o)]) l
    return $ CUnion newl

-- rewriteConcat [FW po, CIntersection [x], FConv(FW o)] = do
--     return $ CConcat [FW po, x, FConv(FW o)]

-- rewriteConcat [FConv(FW o), CUnion l, FW po] =
--     return $ CUnion $ map (\e -> (CConcat [FConv(FW o), e, FW po])) l

rewriteConcat [FI n, CUnion l, FI n'] =
    return $ CUnion $ map (\e -> (CConcat [FI n, e, FI n'])) l

rewriteConcat l =
    if elem F0 l then return F0
     else
         do rl <- mapM relRewrite l
            return $ CConcat rl

rewriteUnion []     = return F0
rewriteUnion [x]     = return x
-- Kill 0
rewriteUnion (F0:r) = rewriteUnion r
rewriteUnion (CUnion l:l1) = 
    return $ CUnion (l++l1)

rewriteUnion (x:l) = do
    x' <- relRewrite x
    return $ CUnion (x':l)

rewriteInter []  = return $ CIntersection []
rewriteInter [x] = do --debug "aqui estamos \n"
    relRewrite x


rewriteInter [FK v1, FPred p] = do
    p' <- relRewrite $ FPred p
    return $ CIntersection [FK v1, p']

rewriteInter [FK v1,FK v2] =
    let u = tryUnify v1 v2 in
            case u of
                   Right v  -> return $ (FK v)
                   Left err -> return $ F0

rewriteInter (FK v1:FK v2:r) =
    let u = tryUnify v1 v2 in
            case u of
                   Right v  -> return $ CIntersection $ (FK v):r
                   Left err -> return $ CIntersection $ F0:r


-- (l1 U l2 U l3) /\ C -> (l1 /\ C) U (l2 /\ C) ...
rewriteInter [(CUnion l), c] =
    return $ CUnion (map (\le -> CIntersection [le, c]) l)

-- C /\ (l1 U l2 U l3) -> (l1 /\ C) U (l2 /\ C) ...
rewriteInter [c, (CUnion l)] =
    return $ CUnion (map (\le -> CIntersection [c, le]) l)

-- A /\ (B /\ C) -> (A /\ B /\ C)
rewriteInter [a, (CIntersection l)] =
    return $ CIntersection $ (a:l)

-- Comment this meta redo out
-- Meta-reductions.

rewriteInter [FK a, CConcat [FI n, e, FI n']] =
    return $ CIntersection $
               [ CConcat [ FI n,
                           CIntersection
                             [ CConcat [ FI n, FK a, FI n ], e ],
                           FI n
                         ], (FK a)
               ]

-- Basic rewriting rule.
rewriteInter [FK a, CConcat [FW p, q, FConv (FW po)]] =
    return $ CConcat
               [ FW p, CIntersection
                 [ CConcat
                   [ FConv (FW po),
                     (FK a),
                     FW p
                   ],
                   q
                 ],
                 FConv (FW po)
               ]

-- Basic rewriting rule.
rewriteInter (F0:l) = return F0
-- This is for predicate fail.
rewriteInter [a,F0] = return F0

-- rewriteInter m@(l@(FK v):l1) = do
--     l1' <- relRewrite $ CIntersection l1
--     return $ CIntersection [l,l1']
--     -- l2' <- rewriteInter [l,l1']
--     -- return l2'

rewriteInter (l:l1:l2:l3) = do
    l' <- rewriteInter [l,l1]
    return $ CIntersection (l':l2:l3)

rewriteInter (l:l1) = do
    l' <- relRewrite l
    return $ CIntersection (l':l1)

generate :: Int -> Int -> [ITerm]
generate 0 _ = []
generate n i = (TVar (i+1)) : generate (n-1) (i+1)

-- IMPORTANT: As we are going to "simulate" unification, this also
-- must rename the variables!
permute :: Vector -> Perm -> Vector
permute v (CPerm l) =
    let m1 = zip l v
        m2 = sort m1
        (_,m3) = unzip m2 in
    fixvars l m3

fixvars :: [Int] -> Vector -> Vector
fixvars l v =
    let ren = renn l in
       map (\t -> fmap ren t) v

renn :: [Int] -> (Int -> Int)
renn l = \n -> if n > 0 then l!!(n-1) else n

freshAndReduce :: Int -> Vector -> Vector
freshAndReduce n l = 
    let v     = take n l
        ren i = if i > n then -i else i in
    map (fmap ren) v
                

-- permute' v p@(CPerm l) =
--     if length v /= maximum l then
--         let needed = (maximum l) - (length v)
--             gen    = generate needed (length v) in
--         permute' (v++gen) p
--     -- XXX Is this right? I guess I'll have to rewrite the engine to figure out.
--     -- if length v /= maximum l then
--     --     permute v (CPerm $ take (length v) l)
--    else
--         let renmap = zip l [1..]
--             ren t = case lookup t renmap of
--                       Nothing -> error $ "Partial permutation " ++ (show t) ++ (show renmap) ++ show p
--                       Just i  -> i 
--             v' = map (\t -> fmap ren t) v in
--         --    trace (show $ (permutev v' p, v, p))  permutev v' p
--         permutev v' p

-- permutev :: Vector -> Perm -> Vector
-- permutev v (CPerm [])     = []
-- permutev v (CPerm (n:nt)) = (v!!(n-1)) : (permutev v (CPerm nt))

-- Start of brain-damaging section:
-- This is a set of non-working hacks to escape from the relational
-- setting to perform unification.

-- The first version was wroten long ago and I don't exactly recall
-- what bug tried to fix using the extra "number" parameter.

-- A simpler one is below, in case we'd need that to check for bugs.

-- Vector to/"a" term.
-- vat n v = TStruct "fake" (freezeCTVars n v)

-- tav (TStruct _ v) = v
-- tav x = [x]

-- freezeCTVars :: Int -> Vector -> [STerm]
-- freezeCTVars n l = map (freezeTerm n) l

-- freezeTerm              :: Int -> ITerm -> STerm
-- freezeTerm i (TVar n)       = TVar ("_" ++(show i)++"x"++(show n))
-- freezeTerm i (TStruct n lt) = TStruct n $ map (freezeTerm i) lt

-- tryUnify :: Vector -> Vector -> Either String Vector
-- tryUnify v1 v2 =
--     let minl = min (length v1) (length v2)
--      v1'  = take minl v1
--      v2'  = take minl v2
--      (nmax, vmax) = if (length v1) > (length v2) then (1, v1) else (2, v2) in
--        do s <- unify (vat 1 v1') (vat 2 v2')
--           return $ tav (apply s (vat nmax vmax))

tryUnify = tryUnify2

-- This is the second and more logical attempt.
tryUnify2 :: Vector -> Vector -> Either String Vector
tryUnify2 v1 v2 =
    let minl = min (length v1) (length v2)
        v1'  = take minl v1
        v2'  = take minl v2
        vmax = if (length v1) > (length v2) then v1 else v2 in
    do s <- unify (vat2 v1') (vat2 v2')
       return $ tav2 (apply s (vat2 vmax))

tav2 (TStruct _ v) = map unVars v

unVars :: STerm -> ITerm
unVars (TVar n)       = TVar (read n)
unVars (TStruct n lt) = TStruct n $ map unVars lt


vat2 v = TStruct "fake" (map tterm v)

tterm :: ITerm -> STerm
tterm (TVar n)   = TVar (show n)
tterm (TStruct n lt) = TStruct n $ map tterm lt
