{-# OPTIONS_GHC -fallow-undecidable-instances #-}
module Engine.Relational where

import AST.Vector
import AST.Prolog
import AST.Relational
import Engine.Rewritable
import Engine.Unification (unify, apply)

import Compiler.Module

import Data.List
import Control.Monad.Trans
import Control.Monad.State
import Control.Monad.Error
import Debug.Trace

type RModule = Module Predicate Relational

class Rew2 a where
    rew :: (a -> Maybe a) -> a -> a

t = (CConcat [FConv (FW (CPerm [])), FK [], FW (CPerm []), F0, F0])

init  a = a 

-- "outermost"
-- XXX: Should we support AC? (If that's the case we are not in very good shape)
rewRel :: (Relational -> Maybe Relational) -> Relational -> Maybe Relational
rewRel r t =
    case r t of 
      Just a  -> Just a
      Nothing ->         
          case t of
            CUnion        l -> maybe Nothing (return . CUnion) (r2 r l)
            CConcat       l -> maybe Nothing (return . CConcat) (r2 r l)
            CIntersection l -> maybe Nothing (return . CIntersection) (r2 r l)
            _               -> Nothing

r2 :: (Relational -> Maybe Relational) -> [Relational] -> Maybe [Relational]
r2 r []     = Nothing
r2 r (t:tt) = 
    case r t of 
      Nothing -> maybe Nothing (return . (t:)) (r2 r tt)
      Just a  -> Just (a:tt)

-- Problem here is that A /\ B /\ C and so on doesn't match B /\ C.
rewRel' :: (Relational -> Maybe Relational) -> Relational -> Relational
rewRel' r t = maybe t id (rewRel r t)
--
r1 F0 = return $ FK []
r1 _  = Nothing

r3 (FK [TVar 1]) = return F0
r3 _  = Nothing

rr2 (CConcat (FConv (FW po):FK v:FW o:x)) = return $ CConcat x
rr2 _                                     = Nothing

-- Rewriting with state
-- We need a function which comes

-- To continue make a list of rewriting rules and apply them in these
-- order. I guess the last one should be instantiation.

ruleP :: Relational -> Maybe Relational
--MonadState (Module Predicate Relational) m =>  Relational -> t m Relational 
ruleP (FPred p) =
    case mlookupCode p of
      (newexpr:_) -> return newexpr
      []          -> return F0
ruleP _ = Nothing

mlookupCode :: Predicate -> [Relational]
mlookupCode p = []

-- This module is at early stage of implementation.
debug s = liftIO $ putStr s

instance (MonadIO m, MonadState RModule m) => Rewritable m Relational where
    isFinal = \_ -> return True
    rewriteStep = return . id

-- Ehem.
rewriteInter (FK v1:FK v2:r) =
    let u = tryUnify v1 v2 in
            case u of
                   Right v  -> return $ CIntersection $ (FK v):r
                   Left err -> return $ CIntersection $ F0:r

-- (l1 U l2 U l3) /\ C -> (l1 /\ C) U (l2 /\ C) ...
-- C /\ (l1 U l2 U l3) -> (l1 /\ C) U (l2 /\ C) ...
-- A /\ (B /\ C) -> (A /\ B /\ C)
-- A /\ I Q I -> I[IAI ∩ Q]I ∩ A
-- A ∩ W Q Wº -> W [Wº A W ∩ Q] Wº

-- Aux functions.
generate :: Int -> Int -> [ITerm]
generate 0 _ = []
generate n i = (TVar (i+1)) : generate (n-1) (i+1)

-- XXX: Move permutation where it belongs.
-- IMPORTANT: As we are going to "simulate" unification, this also
-- must rename the variables!
permute :: Vector -> Perm -> Vector
permute v p@(CPerm l) =
    if length v /= maximum l then
        let needed = (maximum l) - (length v)
            gen    = generate needed (length v) in
        permute (v++gen) p
    -- XXX Is this right? I guess I'll have to rewrite the engine to figure out.
    -- if length v /= maximum l then
    --     permute v (CPerm $ take (length v) l)
    else
        let renmap = zip l [1..]
            ren t = case lookup t renmap of
                      Nothing -> error $ "Partial permutation " ++ (show t) ++ (show renmap) ++ show p
                      Just i  -> i 
            v' = map (\t -> fmap ren t) v in
        --    trace (show $ (permutev v' p, v, p))  permutev v' p
        permutev v' p

permutev :: Vector -> Perm -> Vector
permutev v (CPerm [])     = []
permutev v (CPerm (n:nt)) = (v!!(n-1)) : (permutev v (CPerm nt))

-- Meta-unification
tryUnify :: Vector -> Vector -> Either String Vector
tryUnify v1 v2 =
    let minl = min (length v1) (length v2)
        v1'  = take minl v1
        v2'  = take minl v2
        vmax = if (length v1) > (length v2) then v1 else v2 in
    do s <- unify (v2t v1') (v2t v2')
       return $ t2v (apply s (v2t vmax))

-- ITerm to STerm routines (plus fake vector for terms.
t2v (TStruct "fake" v) = map (fmap read) v
v2t v = TStruct "fake" (map (fmap show) v)
