module LRFact where

-- Left-Right factoring unification algorithm.

import Data.List

-- Not yet.
-- import AST.Ralog

import AST.Prolog

-- Utils for tranlation.
import Translator.String.Prolog
transt s = let Right t = trans_term s
               tn      = fmap (myread) t
               myread ('X':xs) = read xs
           in tn

data T = Cap [T]
       | Conc T T
       | Po Int
       | Funct String Int Int
       | FreeT String
       | Zero
    deriving (Eq)
--    deriving (Eq,Show)

instance Show T where
    show (Cap l)       = "(" ++ (concat $ intersperse " /\\ " (map show l)) ++ ")"
    show (Conc a b)    = (show a) ++ (show b)
    show (Po i)        = "P_"++ (show i)
    show (Funct s i j) = s ++ "_" ++ (show i) ++ "^" ++ (show j)
    show (FreeT s)     = "{"++s++",_}"
    show Zero = "0"

term_t :: ITerm -> T
term_t (TVar i) = Po i
term_t (TStruct  p []) = FreeT p
term_t (TStruct  p t)  =
    let args = map term_t t
        len  = length t
        funz = zipWith (addfun p len) [1..len] args
        addfun p l i t = Conc (Funct p i l) t
    in Cap funz

tt = term_t . transt
ttt t1 t2 = Cap [tt t1, tt t2]

flatten :: T -> T
flatten (Cap l) =
    let caps = filter is_cap l
        caps' = map flatten caps
        c'' = map (\(Cap l) -> l) caps'
    in Cap $ (concat c'') ++ (filter (not.is_cap) l)
flatten l = l

is_cap (Cap l) = True
is_cap _ = False

unify :: T -> T
unify = id

factor_left :: T -> T
factor_left t = repeat_until_eq factor_l t

-- Some note, we should use groupBy/foldr extensively, and then the
-- algorithm will look very nice.

factor_l :: T -> T
factor_l (Cap l) =
    let pairs = cross l
        pairs' = groupBy same_f l
    in Cap $ map common_root pairs
factor_l (Conc a b) = Conc (factor_l a) (factor_l b)
factor_l l = l -- Otherwise do nothing

cross [] = []
cross (x:xs) = [(x,y) | y <- xs, same_f x y && x /= y] ++ cross xs


same_f (Conc a b) d = same_f a d
same_f d (Conc a b)  = same_f d a
same_f f1@(Funct _ _ _) f2@(Funct _ _ _) = f1 == f2
same_f _ _ = False


common_root (t1,t2) =
    let root (Conc a b) = root a
        root a = a
        rest (Conc a b) = b
        rest a = a
    in Conc (root t1) (Cap [rest t1, rest t2])

--     let samefuntor f1@(Funct _ _ _) f2@(Funct _ _ _) = f1 == f2
--         samefuntor (Conc a b) (Conc c d) = samefunctor a c
--         samefuntor _ _ = False
--         groups = groupBy samefunctor l
--         unif_f l =
--             let clean_terms = map
--                 root_f l = se_r (head l)
--                 se_r (Conc a b) = se_r a
-- --                se_r (
--         in Cap $ map unif_f groups


repeat_until_eq :: Eq a => (a -> a) -> a -> a
repeat_until_eq f e =
    let e' = f e in
    if e' == e then
        e'
    else
        repeat_until_eq f e'
