module Pretty.Prolog where

import AST.Prolog

import Text.PrettyPrint

import Compiler.Module (Module, moduleToList)

import Pretty.Pretty

--
instance Show a => Pretty (Module Predicate (Clause a)) where
    pretty _ = concat . prettyProlog

prettyProlog :: Show a => (Module Predicate (Clause a)) -> [String]
prettyProlog fm =
    let lr = moduleToList fm in
             map pretty' lr

pretty' :: Show a => (Predicate, [Clause a]) -> String
pretty' (p, lc) = render $ ppPred p $$ nest 4 (vcat (map ppClause lc)) <> text "\n"

ppPred :: Predicate -> Doc
ppPred (p, n) = text p <> text "/" <> int n

ppClause :: Show a => Clause a -> Doc
ppClause (Clause a l) = ppAtom a <+> text ":-" <+>
                        (hsep (punctuate comma (map ppAtom l))) <+> text "."

ppAtom :: Show a => Atom a -> Doc
-- Ugly hack FIXME
ppAtom (Atom "=" [l1, l2]) = ppTerm l1 <> text "=" <> ppTerm l2
ppAtom (Atom n lt) = text n <> parens (hsep (punctuate comma (map ppTerm lt)))

ppTerm :: Show a => Term a -> Doc
ppTerm (TVar v)         = text $ "_x" ++ show v
ppTerm s@(TStruct "_btlnnil" _)  = text "[]"
ppTerm s@(TStruct "_btlnlist" [a,b])  = text "[" <> ppTerm a <> text "|" <> ppTerm b <> text "]"
ppTerm s@(TStruct _ _)  = ppStruct s

ppConst :: Const -> Doc
ppConst n = text n

ppStruct :: Show a => Term a -> Doc
ppStruct (TStruct n []) = text n
ppStruct (TStruct n lt) = text n <>
                          parens (hsep (punctuate comma (map ppTerm lt)))
