module Pretty.Ralog where


import Text.PrettyPrint

import AST.Prolog
import AST.Vector
import AST.Ralog

import Compiler.Module (Module, moduleToList)

import Pretty.Pretty

--

instance Pretty (Module Predicate Ralog) where
    pretty _ = concat . prettyralog

instance Pretty Ralog where
    pretty _ = render . ppRalog

prettyralog :: (Module Predicate Ralog) -> [String]
prettyralog fm =
    let lr = moduleToList fm in
             map ppRalogPred lr

ppRalogPred :: (Predicate, [Ralog]) -> String
ppRalogPred (p, r) = render $ ppPred' p $$ nest 2 (vcat (map ppRalog r)) <> text "\n"

ppPred' :: Predicate -> Doc
ppPred' (p, n) = text p <> text "/" <> int n

ppRalog :: Ralog -> Doc
ppRalog (CUnion        l) = parens $ vcat (punctuate (text "+") (map ppRalog l))
--ppRalog (CIntersection l) = parens $ vcat (punctuate (text "*") (map ppRalog l))
ppRalog (CIntersection l) = parens $ hsep (punctuate (text "*") (map ppRalog l))
ppRalog (CConcat       l) = parens $ hsep (punctuate (text ";") (map ppRalog l))
ppRalog (COp           o) = ppOp o

ppOp                    :: Op -> Doc
ppOp ID                 = text "id"
ppOp One                = text "1"
ppOp Zero               = text "0"
ppOp TL                 = text "tl"
ppOp HD                 = text "hd"
ppOp (P i)              = text $ "P" ++ (show i)
ppOp (Conv r)           = ppRalog r <> text "^o"
ppOp (Projection t i j) = text $ t ++ "^" ++ (show i) ++ "_" ++ (show j)
ppOp (RelPred s)        = text s
ppOp (Relation l)       = ppRelation l

ppRelation   :: [(String,String)] -> Doc
ppRelation l =
    let rels = map ( \(x,y) -> (text $ x ++ ";" ++ y)) l in
               braces $ hsep $ punctuate comma rels
