module Pretty.RalogLatex where


import Text.PrettyPrint

import AST.Prolog
import AST.Vector
import AST.Ralog

import Compiler.Module (Module, moduleToList)

import Pretty.Pretty

--
instance Pretty Ralog where
    pretty _ = render . ppRalog

ppRalog :: Ralog -> Doc
ppRalog (CUnion        l) = parens $ vcat (punctuate (text "\\cup") (map ppRalog l))
ppRalog (CIntersection l) = parens $ hsep (punctuate (text "\\cap") (map ppRalog l))
--ppRalog (CConcat       l) = parens $ hsep (punctuate (text "") (map ppRalog l))
ppRalog (CConcat       l) = hsep (punctuate (text "") (map ppRalog l))
ppRalog (COp           o) = ppOp o

ppOp                    :: Op -> Doc
ppOp ID                 = text "id"
ppOp One                = text "\\1"
ppOp TL                 = text "tl"
ppOp HD                 = text "hd"
ppOp (P i)              = text $ "P_" ++ (show i)
ppOp (Conv r)           = ppRalog r <> text "\\op"
ppOp (Projection t i j) = text $ t ++ "^" ++ (show i) ++ "_" ++ (show j)
ppOp (RelPred s)        = text s
ppOp (Relation l)       = ppRelation l

ppRelation   :: [(String,String)] -> Doc
ppRelation l =
    let rels = map ( \(x,y) -> (text $ x ++ "," ++ y)) l in
               (text "\\{") <> (hsep $ punctuate comma rels) <> (text "\\}")
