module Pretty.Relational where

import Text.PrettyPrint

import AST.Prolog
import AST.Vector
import AST.Relational

import Compiler.Module (Module, moduleToList)

import Pretty.Pretty

import qualified Pretty.Prolog as PP

--

instance Pretty (Module Predicate Relational) where
    pretty l = concat . (prettyRelational l)

instance Pretty Relational where
    pretty l r = render $ ppRel l r

prettyRelational :: Level -> (Module Predicate Relational) -> [String]
prettyRelational l fm =
    let lr = moduleToList fm in
             map (ppRelPred l) lr

ppRelPred :: Level -> (Predicate, [Relational]) -> String
ppRelPred l (p, r) = render $ ppPred' p $$ nest 2 (vcat (map (ppRel l) r)) <> text "\n"

ppPred' :: Predicate -> Doc
ppPred' (p, n) = text p <> text "/" <> int n

ppRel :: Level -> Relational -> Doc
ppRel v (CUnion l)        = brackets $ vcat (punctuate (text " U") (map (ppRel v) l))
ppRel v (CIntersection l) = brackets $ vcat (punctuate (text " /\\") (map (ppRel v) l))
ppRel v (CConcat l)       = brackets $ hsep (punctuate (text " .") (map (ppRel v) l))

ppRel _ (FK v)    = text "K" <> ppVector v
ppRel v (FW s)    = ppPerm v s
ppRel _ (FPred s) = ppPred' s 
ppRel v (FConv f) = text "cnv" <> parens (ppRel v f)
ppRel _ (FI i)    = text $ "I" ++ show i
ppRel _ (F0)      = text $ "0"

ppPerm :: Level -> Perm -> Doc
ppPerm High (CPerm v) = text "W" <> (parens $ hsep $ punctuate comma (map (text . show) (zip [1..] v)))
ppPerm Low (CPerm _) = text "W"

ppVector :: Vector -> Doc
ppVector l = parens (hsep (punctuate comma (map PP.ppTerm l)))
