module Translator.Prolog.Relational where

import qualified Data.Map as M
-- import Util.StrictMap

import AST.Prolog
import AST.Vector
import AST.Relational

import Compiler.Module (Module, mapModuleList)

import Translator

import Data.List      ( (\\), elemIndex )
import Control.Monad.State

-- For debug purposes.
import Translator.String.Prolog
tt s = let Right t = trans_term s in t
tc s = let Right t = trans_clause s in t

doClause :: Clause String -> (Clause Int, [ITerm])
doClause cl = 
    let (nc, v) = buildVector cl in
    (nc, freshenVector v)
                  
-- Utils for Relational translation
-- This code is really horrible.
buildVector :: Clause String -> (Clause Int, [STerm])
buildVector (Clause head tail) =
    let (nh, v) = headVector head 
        (nt, v') = tailVector v tail in
    (Clause nh nt, v')

makeVar :: a -> Term a
makeVar n = TVar n

headVector :: Atom a -> (Atom Int, [Term a])
headVector (Atom n l) = (Atom n il, l)
    where il = map (makeVar) [1..(length l)]


tailVector :: (Eq a) => [Term a] -> [Atom a] -> ([Atom Int], [Term a])
tailVector v []     = ([], v)
tailVector v (a:at) = 
    let (ai, v') = atomVector v a 
        (al, vl) = tailVector v' at in
    (ai:al, vl)

atomVector :: (Eq a) => [Term a] -> Atom a -> (Atom Int, [Term a])
atomVector v (Atom n l) = (Atom n (map makeVar il), vn)
    where (il, vn) = addLToVector v l

addLToVector :: (Eq a) => [Term a] -> [Term a] -> ([Int], [Term a])
addLToVector v []     = ([], v)
addLToVector v (t:tt) = 
    let (i, v') = addToVector v t 
        (il, vl) = addLToVector v' tt in
    (i:il, vl)

addToVector :: Eq a => [Term a] -> Term a -> (Int, [Term a])
addToVector v t@(TVar a) = 
    case elemIndex t v of
      Nothing -> ((length v) + 1, v++[t])
      Just i  -> ((i+1), v)
addToVector v t@(TStruct _ _) = ((length v) + 1, v++[t])

-- Fresh vectors
freshenVector :: [Term String] -> [Term Int]
freshenVector l = 
    let l' = completeVector l
        ln = zip l' [1..]
        mmap = filter (\(v,n) -> is_var v) ln in
    map (\t -> fmap (subst mmap) t) l'

subst :: [(Term String, Int)] -> String -> Int
subst map x = case lookup (TVar x) map of
            Nothing -> error "Not possible"
            Just i  -> i

completeVector l = 
    let toplevelvars = filter is_var l
        allvars      = map makeVar (vars_list l)
        missing      = allvars \\ toplevelvars in
    l ++ missing

----------------------------------------------------------------
-- Translation, Part II: Structural translation.
----------------------------------------------------------------

-- FIXME: Types are yet very wrong.

instance Translator (Module Predicate (Clause String)) Predicate Relational where
    translate a = Right $ translateProlog a

translateProlog :: Module Predicate (Clause String)->
                   Module Predicate Relational
translateProlog pprog =
    mapModuleList translatePredicate pprog

translatePredicate :: Predicate -> [Clause String] -> [Relational]
translatePredicate p l =
    let np = map doClause l in
             [CUnion $ map translateClause np]

translateSclause :: Clause String -> Relational
translateSclause c = translateClause (doClause c)

-- Head is now normalized.
translateClause :: (Clause Int, [ITerm]) -> Relational
translateClause ((Clause head lt), vector) =
    let innerexpr = (FK vector) : (map (translateAtom (length vector)) lt)
        arity'    = FI (arity head) in
             case innerexpr of
                  -- If innerexpr is a single elem, don't intersect it.
                  [x] -> CConcat $ [arity', x, arity']
                  _   -> CConcat $ [arity', CIntersection innerexpr, arity']

-- Each atom is translated to a W(p) ; atom ; W(p)
translateAtom :: Int -> Atom Int -> Relational
translateAtom i a@(Atom n lt)  =
    let perm = completePerm i lt
        pred = (n, length lt)          in
               CConcat [FW perm, FPred pred, FConv (FW perm)]

-- XXX: This doesn't generate permutations.
completePerm :: Int -> [Term Int] -> Perm 
completePerm n lt = 
    let perm = map (\(TVar i) -> i) lt
        rest = [1..n] \\ perm 
        start = n - length perm in
--    CPerm $ perm ++ (genrest start n rest)
    CPerm $ perm ++ rest

-- genrest :: Int -> Int -> [Int] -> [Int]
-- genrest s n l = zip [(s+1)..n] l
